// Lubachevsky-Stillinger-like algorithm for hard ellipses
// Code is still generating overlaps when it really shouldn't; need to get to the bottom of this
#include"cellmanagement.h"
#include"sched.h"
#include"time.h"
#include"complex.h"
#define PI 3.1415926536
double phi_xtal = PI/(2*sqrt(3)); // density of the monodisperse-ellipse crystal

int d = 2;  // spatial dimension
int N, ninserted;  // # of particles, # of particles inserted so far
double alpha, sigma; // ellipse aspect ratio, semiminor axis
double D; // cell side length
double skin;  // Verlet-list skin
int ncell;  // number of subcells along each side
double lo, hi;
cell supercell;
vector< vector<double> > Rall, Rlast;
vector<double> rtrial;
long long int ntrial;

// For the Marsaglia RNG
int save;
double second;
double *u;
int i97,j97;
double c,cd,cm;
void setup (int seed);
double uniform();

/*Distance of Closest Approach of two arbitrary ellipses*/
// Code adapted from Xiaoyu Zheng's: see http://www.math.kent.edu/~zheng/ellipse.html
extern double distance2d(double a1, double b1, double a2, double b2, double k1dotd, double k2dotd, double k1dotk2);
/* Principal cubic root of a complex number */
extern complex<double> c_cbrt(complex<double> x);

// For SWAP
vector <double> Gi, currentsigmavals, gaps; // gaps = gtilde
vector <int> pindices;
double maxGi;

// for file I/O
ofstream fout1;
void writeconfigs();

// Sorts particles from lowest to highest Gi values
// See  https://stackoverflow.com/questions/25921706/creating-a-vector-of-indices-of-a-sorted-vector
// Same routine is used to identify immobile particles in hdMD
struct IdxCompare2 {
    const std::vector<double>& target;
    IdxCompare2(const std::vector<double>& target): target(target) {}
    bool operator()(int a, int b) const { return target[b] > target[a]; }
};


int main (int argc, const char *argv[]) {
    d = 2;  // spatial dimension
    N = atoi(argv[1]);  // number of ellipses
    alpha = atof(argv[2]);  // particle aspect ratio
    int mrd = atoi(argv[3]);
    double maxrotdisp = (PI/180)*(mrd/alpha);   //(PI/18)/alpha;
    int swappct = atoi(argv[4]);  // Attempt swaps for this percentage of particles
    bool TDOFflag = atoi(argv[5]); // Perform TDOF moves only if TDOFflag = true;
    int trialnumber = atoi(argv[6]);
    D = sqrt(N*alpha); // side length given in the JPCB article
    // sqrt((PI/4)*N*alpha/phi_xtal);  // This side length makes monodisperse ellipses packed at phi = phi_xtal have unit-length minor axes

    const long long int maxtrialsperparticle = 25000000;  // This should always be large enough for systems to jam
    const long long int maxtrials = N*maxtrialsperparticle;

    chrono::duration<double> elapsedtime;
    auto start = chrono::high_resolution_clock::now();;
    
// Declare counters, position vectors, etc.
    int i, j, k, q, r;
    double dum;
    long long int nsucc, nsucclast; // number of successful particle moves, swap attempts/successful moves
    nsucc = 0;  nsucclast = 0;

    bool succmoveflag;
    Rlast.resize(N);
    for (i = 0; i < N; i++) Rlast[i].resize(d+2);
    rtrial.resize(d+2);
    double vol = pow(D,d);
    double totpartvol, phi, phi_last;
    lo = -D/2;  hi = D/2;  // set supercell boundaries
    double DSR; // MC move success rate SINCE THE PREVIOUS LOG-FILE ENTRY
// Extra variables for SWAP MC
    long long int swapatt, swapsucc, lastswapatt, lastswapsucc; // number of swap attempts/successful moves
    swapatt = 0;    swapsucc = 0;   lastswapatt = 0;    lastswapatt = 0;
    bool goodj; // added 8/15/24
    int attthisj; // added 8/15/24
    
    Gi.resize(N);    gaps.resize(N);    currentsigmavals.resize(N);
    double sigtmp;
    int qmax = round((swappct/100.0)*N);
    double DSSR; // SWAP move success rate SINCE THE PREVIOUS LOG-FILE ENTRY
    bool redmaxdisp = false; // becomes true one maximum translational/rotational displacements start dropping
// Set up the linked subcells
    double currentmaxsigma = 0;
    double initmaxsigma = .07;
    double ratrand, prand;
    skin = .5;
    ncell = floor(D/(2*initmaxsigma*alpha+skin)); // This is the most efficient value of ncell
    supercell.setupsupercell();
    int ncell_last = ncell;

// Seed random number generator
    unsigned int sseed = time(NULL)-6*trialnumber + trialnumber*trialnumber;
    srand(sseed);
    setup((rand()+trialnumber)%899999998 + 1);

// Set frequency of log-file output
    int nlog = N;

    cout << "Rob Hoy's jammed-ellipse-packing-generation code" << endl;
    cout << "If you use this to generate anything that is included in a publication, please cite: R. S. Hoy, J. Phys. Chem. B 129, 763 (2025)." << endl;
    
// Generate the initial configuration
    ninserted = 0;  ntrial = 0;
    do {
        for (j = 0; j < d; j++) rtrial[j] = (uniform()-0.5)*D;
        rtrial[d] = uniform()*PI;

// Polydisperse systems: P(σ) ~ σ^-2 with max ratio 1.4; this is a continuous analog of the std. Corey/Andrea model, but with equal volume fractions of particles with different diameters
        do {
            ratrand = 5.0/7 + uniform()*2.0/7;
            rtrial[d+1] = ratrand*initmaxsigma;
            prand = uniform();
            } while (prand > .510204/(ratrand*ratrand));
        if (supercell.doesitfit(ninserted,true)) {
            totpartvol += PI*alpha*Rall[ninserted-1][3]*Rall[ninserted-1][3];
            phi = totpartvol/vol;
        }
// Track the run's progress
        if (!(ntrial%nlog))  {
            auto curr = chrono::high_resolution_clock::now();
            elapsedtime = curr-start;
            cout << "Trials/(unit volume) = " << ntrial/vol << "  ninserted = " << ninserted << "  phi = " << phi << ", t = " << elapsedtime.count() << endl;
        }
        ntrial++;
    } while (ninserted != N);
    phi_last = phi;
    cout << "Initial state created." << endl;
    
// File I/O -- some of this can be dropped in our app
    char configname[80], logfname[80];
// Log file recording the run's progress
    ofstream fout2;
    if (TDOFflag) sprintf(logfname, "log.L%.2f.alpha%.3f.swappct.%d.maxrot.%d.withTDOF.trial%d",D,alpha,swappct,mrd,trialnumber);
    else sprintf(logfname, "log.L%.2f.alpha%.3f.swappct.%d.maxrot.%d.noTDOF.trial%d",D,alpha,swappct,mrd,trialnumber);
    fout2.open(logfname);

// Set the Monte Carlo parameters
    double maxtransdisp = .05;
    double sizefac = 1;
    double mingapdist, attGi;
    char logoutput[250];

// Implement our Lubachevsky-Stillinger-like particle-growth algorithm
// This method ALMOST preserves shape of P(σ), ratios of all moments of P(σ)
    nsucc = 0;
    int nconsec = 0;
    for (ntrial = 1; ntrial < maxtrialsperparticle; ntrial++) {
// Set frequency of log-file output
        if (ntrial <= 1000) nlog = 10;
        else if (ntrial <= 100000) nlog = 100;
        else nlog = 1000;
// Rebuild and repopulate the subcells -- need do this differently for large alpha???
        for (q = 0; q < N; q++) if (Rall[q][3] > currentmaxsigma) currentmaxsigma = Rall[q][3];
        ncell = floor(D/(2*currentmaxsigma*alpha+skin)); // This is the most efficient value of ncell
        if (ncell != ncell_last) {
            supercell.setupsupercell();
            ncell_last = ncell;
        }
        supercell.populatecells();
        
// STAGE 1: Attempt to translate each particle by up to sizefac*maxtransdisp along each spatial dimension, rotate each particle by up to sizefac*maxrotdisp degrees
// See the below sizefac-reduction criterion!
        succmoveflag = false;
        for (q = 0; q < N; q++) {
// Generate the trial move
            for (j = 0; j < d; j++) {
                rtrial[j] = Rall[q][j] + (uniform()-0.5)*sizefac*maxtransdisp;
                if (rtrial[j] < -D/2) rtrial[j] += D;       // new position must be in the box
                else if (rtrial[j] > D/2) rtrial[j] -= D;
            }
            rtrial[2] = Rall[q][2] + (uniform()-0.5)*sizefac*maxrotdisp;
            rtrial[3] = Rall[q][3];
// supercell.doesitfit() = true if the particle can be inserted, otherwise false
            if (supercell.doesitfit(q,false)) {
                for (j = 0; j < d+2; j++) Rall[q][j] = rtrial[j];
                nsucc++;    succmoveflag = true;
            }
        }
        
// STAGE 2: Grow the particles by a factor \tilde{\mathcal{G}}
// Here diamreductionflag = true
        maxGi = supercell.findmingapdist(1)/(2*alpha);
        if (maxGi < 0) { cout << "Error: Gtilde < 0" << endl; exit(1);  }
        if (maxGi > 0) for (q = 0; q < N; q++) Rall[q][3] *= 1 + min(maxGi,.001);
        
// STAGE 3: SWAP particles with lower Gi with smaller particle that have higher Gi, provided the size difference isn't too large
// Again, here diamreductionflag = true
// ONLY DO THIS IF SWAPPCT > 0
        supercell.findmingapdist(1);
        pindices.clear();
        for (q = 0; q < N; ++q) pindices.push_back(q);
        sort(pindices.begin(),pindices.end(),IdxCompare2(gaps));
// Swap all particles with a randomly chosen smaller particle with a larger gap IF that particle will fit
// gaps[pindices[j]] > gaps[pindices[q]] by construction, now we need to find a particle j with a smaller diameter
// Keep selecting j values until a smaller particle has been found or 100 j values have been tried
        if (swappct) for (q = 0; q < qmax; q++) {
            goodj = false;
            attthisj = 0;
            do {
                j = round(q + 1 + (N-2-q)*(1.0*rand())/RAND_MAX);
                attthisj++;
                if (Rall[pindices[j]][3] < Rall[pindices[q]][3]) {
                    goodj = true;
                    if (gaps[pindices[j]] - gaps[pindices[q]] > Rall[pindices[q]][3] - Rall[pindices[j]][3]) {
                        for (k = 0; k < 3; k++) rtrial[k] = Rall[pindices[q]][k];
                        rtrial[3] = Rall[pindices[j]][3];
                        if (supercell.doesitfit(pindices[q],false)) {
                            for (k = 0; k < 3; k++) rtrial[k] = Rall[pindices[j]][k];
                            rtrial[3] = Rall[pindices[q]][3];
                            if (supercell.doesitfit(pindices[j],false)) {
                                sigtmp = Rall[pindices[j]][3];
                                Rall[pindices[j]][3] = Rall[pindices[q]][3];
                                Rall[pindices[q]][3] = sigtmp;
                                supercell.getlocalmingapdist(pindices[j],1,1);
                                supercell.getlocalmingapdist(pindices[q],1,1);
                                swapsucc++;
// Reorder the particles in order of increasing gaps, using the new gaps values for particles q, j and their nearest neighbors
                                pindices.clear();
                                for (r = 0; r < N; ++r) pindices.push_back(r);
                                sort(pindices.begin(),pindices.end(),IdxCompare2(gaps));
                            }
                        }
                    }
                }
            } while (!goodj && attthisj < N/10);
            swapatt++;
        }

// Reduce MC trial-move sizes when success rate gets too low
// End the run if maxtransdisp drops below 10^-10
        maxGi = supercell.findmingapdist(1)/(2*alpha);
        if ((!redmaxdisp && maxGi < pow(10.0,-9.0)) || (redmaxdisp &&  maxGi < pow(10.0,-15))) nconsec++;
        else nconsec = 0;
        if (ntrial > 1 && nconsec > 99) {
            sizefac *= .75;
            cout << "Decreasing MC trial-move sizes: new sizefac = " << sizefac << " maxtransdisp = " << sizefac*maxtransdisp << " maxrotdisp = " << sizefac*maxrotdisp << endl;
            nconsec = 0;
            if (sizefac < 2*pow(10.0,-7)) redmaxdisp = true; // Should try adjusting this cutoff in LSps-v12
            if (sizefac*maxtransdisp < pow(10.0,-10)) break;
        }
        
// STAGE 4: Late in the run, grow particles by q-dependent values
// ONLY IF TDOFFLAG = TRUE
        if (TDOFflag && sizefac < .01) {
            supercell.findmingapdist(1);
            for (q = 0; q < N; q++) Rall[q][3] *= 1 + min(Gi[q]/(2*alpha),.001);
        }
// Store the current particle positions, orientations, diameters
        for (q = 0; q < N; q++) for (j = 0; j < 4; j++) Rlast[q][j] = Rall[q][j];
        
// Track the run's progress
        totpartvol = 0;
        for (i = 0; i < N; i++) totpartvol += PI*alpha*Rall[i][3]*Rall[i][3];
        phi = totpartvol/vol;
        auto curr = chrono::high_resolution_clock::now();
        elapsedtime = curr-start;
// OUTPUT RUN'S PROGRESS TO THE LOG FILE
        if (ntrial == 1 || !(ntrial%nlog)) {
            DSR = (1.0*(nsucc-nsucclast))/(nlog*N);
 	        if (swapatt-lastswapatt > 0) DSSR = (1.0*(swapsucc-lastswapsucc))/(swapatt-lastswapatt);
	        else DSSR = 0;
            sprintf(logoutput,"Growth cycle = %lld     phi = %f     Gtilde = %.6e     nsucc = %lld     DSR = %.3f     nconsec = %d     swapatt = %lld     swapsucc = %lld     DSSR = %.3f     t = %f",ntrial,phi,maxGi,nsucc,DSR,nconsec,swapatt,swapsucc,DSSR,elapsedtime.count());
            cout << logoutput << endl;
            fout2 << ntrial << " " << phi << " " << maxGi << " " << DSR << " " << nconsec << " " << swapatt << " " << swapsucc << " " << DSSR << " " << elapsedtime.count() << endl;
            nsucclast = nsucc;      lastswapsucc = swapsucc;    lastswapatt = swapatt;
        }

} // END OF THE MAIN LOOP
    
// WRAPPING UP: Output the final log data and configurations
// Execute stage 4 three more times to eliminate the remaining gaps IF TDOFflag = TRUE
    if (TDOFflag) {
        supercell.findmingapdist(1);
        for (q = 0; q < N; q++) if (Gi[q]> 0) Rall[q][3] += Gi[q]/(2*alpha);
        supercell.findmingapdist(1);
        for (q = 0; q < N; q++) if (Gi[q]> 0) Rall[q][3] += Gi[q]/(2*alpha);
        supercell.findmingapdist(1);
        for (q = 0; q < N; q++) if (Gi[q]> 0) Rall[q][3] += Gi[q]/(2*alpha);
        supercell.findmingapdist(1); // This will eliminate any tiny overlaps in the final state
    }
    totpartvol = 0;
    for (i = 0; i < N; i++) totpartvol += PI*alpha*Rall[i][3]*Rall[i][3];
    phi = totpartvol/vol;
    DSR = (1.0*(nsucc-nsucclast))/(nlog*N);
    if (swapatt-lastswapatt > 0) DSSR = (1.0*(swapsucc-lastswapsucc))/(swapatt-lastswapatt);
    else DSSR = 0;
    sprintf(logoutput,"Trials/particle = %lld     phi = %f     max. sigma increase = %.6e     nsucc = %lld     DSR = %.3f     nconsec = %d     swapatt = %lld     swapsucc = %lld     DSSR = %.3f     t = %f",ntrial,phi,maxGi,nsucc,DSR,nconsec,swapatt,swapsucc,DSSR,elapsedtime.count());
    cout << logoutput << endl;
    fout2 << ntrial << " " << phi << " " << maxGi << " " << DSR << " " << nconsec << " " << swapatt << " " << swapsucc << " " << DSSR << " " << elapsedtime.count() << endl;
    fout2.close();


// Write the final gaps (tilde g__i) values
    pindices.clear();
    for (r = 0; r < N; ++r) pindices.push_back(r);
    sort(pindices.begin(),pindices.end(),IdxCompare2(gaps));
    char gapfname[80];
    if (TDOFflag) sprintf(gapfname,"finalGifacs.L%.2f.alpha%.3f.swappct.%d.maxrot.%d.withTDOF.trial%d",D,alpha,swappct,mrd,trialnumber);
   else sprintf(gapfname,"finalGifacs.L%.2f.alpha%.3f.swappct.%d.maxrot.%d.noTDOF.trial%d",D,alpha,swappct,mrd,trialnumber);
    ofstream fout3;
    fout3.open(gapfname);
    for (i = 0; i < N; i++) fout3 << gaps[pindices[i]] << " ";
    cout << endl;
    fout3.close();

    
// OUTPUT THE FINAL CONFIGURATIONS
// First sort the particles from small to large so these files are easier to read
        for (q = 0; q < N; q++) currentsigmavals[q] = Rall[q][3];
        pindices.clear();
        for (q = 0; q < N; ++q) pindices.push_back(q);
        sort(pindices.begin(),pindices.end(),IdxCompare2(currentsigmavals));
// Now write the final configs.
        char finalconfname[80];
        if (TDOFflag) sprintf(finalconfname, "finalconfigs.L%.2f.alpha%.3f.swappct.%d.maxrot.%d.withTDOF.trial%d",D,alpha,swappct,mrd,trialnumber);
        else sprintf(finalconfname, "finalconfigs.L%.2f.alpha%.3f.swappct.%d.maxrot.%d.noTDOF.trial%d",D,alpha,swappct,mrd,trialnumber);
        fout3.open(finalconfname);
        fout3 << setprecision(10) << fixed;
        fout3 << D << endl;
        for (i = 0; i < N; i++) {
            fout3 << i;
            for (j = 0; j < d+2; j++) fout3 << " " << Rall[pindices[i]][j];
            fout3 << endl;
        }
        fout3.close();
// Scale the final configs. so the smallest particles have sigma = .5
    double sigsmall = 10;
    for (i = 0; i < N; i++) if (Rall[i][3] < sigsmall) sigsmall = Rall[i][3];
    D *= .5/sigsmall;
    for (i = 0; i < N; i++) {
        Rall[i][0] *= .5/sigsmall;      Rall[i][1] *= .5/sigsmall;      Rall[i][3] *= .5/sigsmall;
    }
// Now write the scaled/sorted configs.
    if (TDOFflag) sprintf(finalconfname, "finalscaledconfigs.L%.2f.alpha%.3f.swappct.%d.maxrot.%d.withTDOF.trial%d",D,alpha,swappct,mrd,trialnumber);
    else sprintf(finalconfname, "finalscaledconfigs.L%.2f.alpha%.3f.swappct.%d.maxrot.%d.noTDOF.trial%d",D,alpha,swappct,mrd,trialnumber);
    fout3.open(finalconfname);
    fout3 << setprecision(10) << fixed;
    fout3 << D << endl;
    for (i = 0; i < N; i++) {
        fout3 << i;
        for (j = 0; j < d+2; j++) fout3 << " " << Rall[pindices[i]][j];
        fout3 << endl;
    }
    fout3.close();

// Clear the arrays, close the remaining open file
    Rall.clear();
    Gi.clear();
    rtrial.clear();
    fout1.close();
    
    return 0;
}



// Write ellipse configurations to a file
// Goal is to replace this with a routine that draws them on the screen
void writeconfigs() {
    int i, j;
    fout1 << D << endl;
    for (i = 0; i < N; i++) {
        fout1 << i;
        for (j = 0; j < d+2; j++) fout1 << " " << Rall[i][j];
        fout1 << endl;
    }
    return;
}



/*Distance of Closest Approach of two arbitrary ellipses*/
// Code adapted from Xiaoyu Zheng's: see http://www.math.kent.edu/~zheng/ellipse.html
double distance2d(double a1, double b1, double a2, double b2, double k1dotd, double k2dotd, double k1dotk2) {
     double eps1,eps2,nu,Ap[2][2],lambdaplus,lambdaminus,bp2,ap2,cosphi,tanphi2,delta,dp,qur;
     complex<double> A,B,C,D,E,alphaXZ,beta,gamma,P,Q,U,y,qu;
// XZ's fix from July 2012
     if (k1dotk2 == -1) k1dotk2 = 1;
     eps1=sqrt(1.0-(b1*b1)/(a1*a1));
     eps2=sqrt(1.0-(b2*b2)/(a2*a2));
     nu=a1/b1-1.0;
     Ap[0][0]=b1*b1/(b2*b2)*(1.0+0.5*(1.0+k1dotk2)*
              (nu*(2.0+nu)-eps2*eps2*(1.0+nu*k1dotk2)*(1.0+nu*k1dotk2)));
     Ap[1][1]=b1*b1/(b2*b2)*(1.0+0.5*(1.0-k1dotk2)*
              (nu*(2.0+nu)-eps2*eps2*(1.0-nu*k1dotk2)*(1.0-nu*k1dotk2)));
     Ap[0][1]=b1*b1/(b2*b2)*0.5*sqrt(1.0-k1dotk2*k1dotk2)*
              (nu*(2.0+nu)+eps2*eps2*(1.0-nu*nu*k1dotk2*k1dotk2));
     lambdaplus=0.5*(Ap[0][0]+Ap[1][1])+sqrt(0.25*(Ap[0][0]-Ap[1][1])*
              (Ap[0][0]-Ap[1][1])+Ap[0][1]*Ap[0][1]);
     lambdaminus=0.5*(Ap[0][0]+Ap[1][1])-sqrt(0.25*(Ap[0][0]-Ap[1][1])*
              (Ap[0][0]-Ap[1][1])+Ap[0][1]*Ap[0][1]);
     bp2=1.0/sqrt(lambdaplus);
     ap2=1.0/sqrt(lambdaminus);
     if (fabs(k1dotk2)==1.0) {
        if(Ap[0][0]>Ap[1][1]) cosphi=b1/a1*k1dotd/sqrt(1.0-eps1*eps1*k1dotd*k1dotd);
        else cosphi=sqrt(1.0-k1dotd*k1dotd)/sqrt(1.0-eps1*eps1*k1dotd*k1dotd);
     }
     else {
        cosphi=1.0/sqrt(2.0*(Ap[0][1]*Ap[0][1]+(lambdaplus-Ap[0][0])*(lambdaplus-Ap[0][0]))*
              (1.0-eps1*eps1*k1dotd*k1dotd))*(Ap[0][1]/sqrt(1.0+k1dotk2)*(b1/a1*k1dotd+
              k2dotd+(b1/a1-1.0)*k1dotd*k1dotk2)+ (lambdaplus-Ap[0][0])/sqrt(1.0-k1dotk2)*
              (b1/a1*k1dotd-k2dotd-(b1/a1-1.0)*k1dotd*k1dotk2));
     }
     delta=ap2*ap2/(bp2*bp2)-1.0;
     if (delta==0.0 || cosphi==0.0) dp=1.0+ap2;
     else {
        tanphi2=1.0/(cosphi*cosphi)-1.0;
        A=-(1.0+tanphi2)/(bp2*bp2);
        B=-2.0*(1.0+tanphi2+delta)/bp2;
        C=-tanphi2-(1.0+delta)*(1.0+delta)+(1.0+(1.0+delta)*tanphi2)/(bp2*bp2);
        D=2.0*(1.0+tanphi2)*(1.0+delta)/bp2;
        E=(1.0+tanphi2+delta)*(1.0+delta);
        alphaXZ=-3.0*B*B/(8.0*A*A)+C/A;
        beta=B*B*B/(8.0*A*A*A)-B*C/(2.0*A*A)+D/A;
        gamma=-3.0*B*B*B*B/(256.0*A*A*A*A)+C*B*B/(16.0*A*A*A)-B*D/(4.0*A*A)+E/A;
        if (beta==0.0) {
           qu=-B/(4.0*A)+sqrt(0.5*(-alphaXZ+sqrt(alphaXZ*alphaXZ-4.0*gamma)));
        }
        else {
           P=-alphaXZ*alphaXZ/12.0-gamma;
           Q=-alphaXZ*alphaXZ*alphaXZ/108.0+alphaXZ*gamma/3.0-beta*beta/8.0;
           U=c_cbrt(-Q*0.5+sqrt(Q*Q*0.25+P*P*P/27.0));
           if (U==0.0) y=-5.0*alphaXZ/6.0-c_cbrt(Q);
           else y=-5.0*alphaXZ/6.0+U-P/(3.0*U);
           qu=-B/(4.0*A)+0.5*(sqrt(alphaXZ+2.0*y)+ sqrt(-(3.0*alphaXZ+2.0*y+2.0*beta/sqrt(alphaXZ+2.0*y))));
        }
        qur = real(qu);
        dp= sqrt((qur*qur-1.0)/delta*(1.0+bp2*(1.0+delta)/qur)*(1.0+bp2*(1.0+delta)/qur)+(1.0-(qur*qur-1.0)/delta)* (1.0+bp2/qur)*(1.0+bp2/qur));
     }
     return dp*b1/sqrt(1.0-eps1*eps1*k1dotd*k1dotd);
}


/* Principal cubic root of a complex number */
// Code adapted from Xiaoyu Zheng's: see http://www.math.kent.edu/~zheng/ellipse.html
complex<double> c_cbrt(complex<double> x) {
    double a,b,r,phi,rn;
    a = 1;
    const complex<double> II(0.0,1.0);
    a=real(x);
    b=imag(x);
    r=sqrt(a*a+b*b);
    phi=atan2(b,a);
    phi/=3.0;
    rn=cbrt(r);
    return rn*cos(phi) + II*rn*sin(phi);
}


    
// Set up and seed the Marsaglia RNG
// Code adopted from that used in LAMMPS (https://www.lammps.org/)
void setup(int seed) {
   int ij,kl,i,j,k,l,ii,jj,m;
   double s,t;
   if (seed <= 0 || seed > 900000000) {
       cout << "Invalid seed for Marsaglia random # generator; must be between 1 and 900 million." << endl;
       exit(1);
    }
  save = 0;
  u = new double[97+1];
  ij = (seed-1)/30082;
  kl = (seed-1) - 30082*ij;
  i = (ij/177) % 177 + 2;
  j = ij %177 + 2;
  k = (kl/169) % 178 + 1;
  l = kl % 169;
  for (ii = 1; ii <= 97; ii++) {
    s = 0.0;
    t = 0.5;
    for (jj = 1; jj <= 24; jj++) {
      m = ((i*j) % 179)*k % 179;
      i = j;
      j = k;
      k = m;
      l = (53*l+1) % 169;
      if ((l*m) % 64 >= 32) s = s + t;
      t = 0.5*t;
    }
    u[ii] = s;
  }
  c = 362436.0 / 16777216.0;
  cd = 7654321.0 / 16777216.0;
  cm = 16777213.0 / 16777216.0;
  i97 = 97;
  j97 = 33;
  uniform();
}



// Generate a random # between 0 and 1
// Code adopted from that used in LAMMPS (https://www.lammps.org/)
double uniform() {
  double uni = u[i97] - u[j97];
  if (uni < 0.0) uni += 1.0;
  u[i97] = uni;
  i97--;
  if (i97 == 0) i97 = 97;
  j97--;
  if (j97 == 0) j97 = 97;
  c -= cd;
  if (c < 0.0) c += cm;
  uni -= c;
  if (uni < 0.0) uni += 1.0;
  return uni;
}



